// -------------------------------------------------------------
//  Software für Luftentfeuchter-Steuerung eines Comedes LTR600
//  mit Arduino Micro - ersetzt fehlerhafte Original-Steuerung
//
//  Modul zur Relais-Ansteuerung
//  Juli/Dezember 2014
// -------------------------------------------------------------

#include <stdio.h>
#include <Arduino.h>

#include "thermometer.h"
#include "compressor.h"


// Pinning
// -------
const int PIN_NTC        = A7;      // Connected to NTC / 8k2 voltage divider
const int PIN_COMPRESSOR = 12;      // Transistor witching the compressor relay
const int PIN_LED        = 13;      // Onboard LED


// Main operating parameters
// -------------------------

// All temperatures in degrees Celcius, all times and delays in seconds. Not
// included in the delays are the time for measuring the temperature and the
// code execution time. They both normally can be neglected.
const float        TEMP_CRITICAL_RANGE_LOW       =   1.0;
const float        TEMP_CRITICAL_RANGE_HIGH      =   5.0;
const unsigned int DELAY_CHECK_TEMPERATURE_RAISE =    60;
const unsigned int MAX_EXTRA_TIME                = 20*60;
const unsigned int SLOPE_SHORTEN_EXTRA_TIME      = 10*60; // A temperature raise of 1 degree during DELAY_CHECK_TEMPERATURE_RAISE shortens extra wait time by this amount.
const unsigned int MAX_TIME_BELOW_CRITICAL_RANGE = 20*60;
const unsigned int DEFAULT_LOOP_DELAY            =     5; // Default wait time when inside a loop

#define DEBUG_OUPUT

// ------------------
//  Global instances
// ------------------

t_Thermometer Thermometer (PIN_NTC);
t_Compressor  Compressor  (PIN_COMPRESSOR);

// ----------------
//  Main functions
// ----------------

static int uart_putchar (char c, FILE *stream)
{
   #ifdef DEBUG_OUPUT
      Serial.write(c);
   #endif
   return 0 ;
}

void setup()
{
   pinMode(PIN_LED, OUTPUT);     
   #ifdef DEBUG_OUPUT
      static FILE uartout = {0};

      Serial.begin(9600);
      fdev_setup_stream (&uartout, uart_putchar, NULL, _FDEV_SETUP_WRITE);
      stdout = &uartout;
      printf ("Powered on\n");
   #endif
}

static void WaitSeconds (int Seconds, bool Countdown=true)
{
   if (Seconds > DEFAULT_LOOP_DELAY) // No countdown for default loop delays
   {
      int i = 0;
      while (Seconds >= DEFAULT_LOOP_DELAY)
      {
         if (Countdown)
            printf ("%d...", Seconds);
         if ((++i % 10) == 0)
            printf ("\n");
         delay (DEFAULT_LOOP_DELAY*1000UL);
         Seconds -= DEFAULT_LOOP_DELAY;
      }
   }
   delay (Seconds*1000UL);
}

static void Halt (void)
{
   printf ("Switching compressor off\n");
   Compressor.Off();                      // Safe state
   for (;;)
   {
      printf ("System halted\n");
      WaitSeconds (DEFAULT_LOOP_DELAY);
   }
}

static float GetTemp (void)
{
   float Temp;
   bool  Error;

   Temp = Thermometer.GetAverageTemperature (Error);
   if (Error)
   {
      printf ("ERROR: Temperature sensor not working\n");
      printf ("Please repair...\n");
      Halt();  // There's no sense in continuing with a broken temperature sensor.
   }
   return Temp;
}

static void PrintTemp (float Temp)
{
   #ifdef DEBUG_OUPUT
   Serial.print(Temp,1);   
   printf ("%cC",0xB0); //Might need to be adapted depending on the platform used
   #endif
}


void MainCycle (void)
{
   bool  BadTemp;
   float Temp;
   float Delta;
   float PrevTemp;
   bool  Warmer;

   Temp = GetTemp();
   printf ("Current temperature: ");
   PrintTemp (Temp); 
   printf ("\n");
   if (Temp > TEMP_CRITICAL_RANGE_HIGH)
   {
      printf ("Temperature lies above critical range - no extra wait time.\n");
   }
   else 
   {
      while (Temp < TEMP_CRITICAL_RANGE_LOW)
      {
         Temp = GetTemp();
         printf ("Waiting until temperature >= lowest limit ("); 
         PrintTemp (TEMP_CRITICAL_RANGE_LOW); 
         printf (") -- Current temperature: ");
         PrintTemp (Temp); 
         printf ("\n");
         WaitSeconds (DEFAULT_LOOP_DELAY);
      }

      // The sensor's temperature now is above the lowest limit, but what about the rest of
      // the device? To be really sure that there's no ice let's wait some extra time. The
      // extra wait time depends on how fast temperature goes up. Do not wait at all if the
      // temperature already lies above the critical range.

      printf ("Temperature lies within critical range. Checking how fast temperature raises\n");
      do
      {
         PrevTemp = Temp;
         printf ("Waiting for %d seconds...\n", DELAY_CHECK_TEMPERATURE_RAISE);
         WaitSeconds (DELAY_CHECK_TEMPERATURE_RAISE);
         printf ("\n");
         Temp  = GetTemp();
         Delta = (Temp - PrevTemp);
         printf ("Previous temperature: "); PrintTemp (PrevTemp);
         printf ("\nCurrent  temperature: "); PrintTemp (Temp);
         Warmer = (Delta >= 0.05);
         if (!Warmer)
            printf (" -- Temperature not going up! Waiting some more time...");
         printf ("\n");
      } while (!Warmer);

      if ((Delta * SLOPE_SHORTEN_EXTRA_TIME) < MAX_EXTRA_TIME)
      {
         int ExtraWaitSeconds = MAX_EXTRA_TIME - Delta * SLOPE_SHORTEN_EXTRA_TIME;
         printf ("Waiting for %d extra seconds for device to unfreeze\n", ExtraWaitSeconds);
         WaitSeconds (ExtraWaitSeconds);
      }
      else
      {
         printf ("No extra wait time\n");
      }
   }

   // Device should be free of ice at this point.
   // Let's switch the compressor on - and keep an eye on the temperature.
   // Temperature is allowed to drop below the critical range, but only for
   // certain amount of time

   printf ("Device should be free of ice. Let's switch it on!\n");
   Compressor.On();
   int TotalSecondsBelowCriticalRange = 0;
   do
   {
      Temp = GetTemp();
      printf ("Compressor is on. Current temperature: ");
      PrintTemp (Temp);
      printf ("\n");
      if (Temp < TEMP_CRITICAL_RANGE_LOW)
      {
         printf ("Below critical range! May continue for %d more seconds\n", MAX_TIME_BELOW_CRITICAL_RANGE - TotalSecondsBelowCriticalRange);
         TotalSecondsBelowCriticalRange += DEFAULT_LOOP_DELAY;
      }
      else 
      {
         TotalSecondsBelowCriticalRange = 0;
      }
      WaitSeconds (DEFAULT_LOOP_DELAY);
   } while (TotalSecondsBelowCriticalRange < MAX_TIME_BELOW_CRITICAL_RANGE);
   printf ("Too cold for too long time, switching compressor off.\n");
   Compressor.Off();
}

void loop()
{
   Compressor.Off();
   for (int i=5; i>=0; i--)   // I blink, therefore I am
   {   
      digitalWrite (PIN_LED, HIGH); 
      delay (500);              
      digitalWrite (PIN_LED, LOW);  
      delay (500);              
      printf ("Blink %d\n", i);
   }   
   printf ("int       is %d\n", sizeof(int));
   printf ("long      is %d\n", sizeof(long));
   printf ("long long is %d\n", sizeof(long long));
   printf ("float     is %d\n", sizeof(float));
   printf ("double    is %d\n", sizeof(double));
      
   printf ("Starting\n");
   printf ("Compressor switched off - entering main cylce\n");
   for (;;)
   {
      printf ("Entering main cylce\n");
      MainCycle();
   }
}

