// -------------------------------------------------------------
//  Software für Luftentfeuchter-Steuerung eines Comedes LTR600
//  mit Arduino Micro - ersetzt fehlerhafte Original-Steuerung
//
//  Modul zur Temperatur-Abfrage
//  Basierend auf mehreren Mess-Stuetzen werden die Zwischenräume
//  linear interpoliert. Mehrere Messungen werden gemittelt.
//  Juli/Dezember 2014
// -------------------------------------------------------------

#include <float.h>

typedef struct
{
   int   RawValue;
   float Temp;
   float Slope;
   float Offset;
} t_Sample, *t_pSample;

class t_Thermometer
{
   private:
      const int SamplesPerMeasure   = 10;   // Every measure is the average of this number of samples
      const int DelayBetweenSamples = 10;   // Wait time between individual samples in ms

   private:
      void AddSample (int RawValue, float Temp)
      {
         t_pSample pSample = &oSampleArr[oSamples];
         pSample->RawValue = RawValue;
         pSample->Temp     = Temp;
         if (oSamples > 0)
         {
            t_pSample pPrevSample = &oSampleArr[oSamples-1];
            pSample->Slope  = (pSample->Temp - pPrevSample->Temp) / (pSample->RawValue - pPrevSample->RawValue);
            pSample->Offset = pSample->Temp - (pSample->Slope * pSample->RawValue);
         }
         oSamples++;
      }

   public:
      t_Thermometer (int Pin):
         oPin (Pin),
         oSamples (0)
      {                           // Refrence measures that have been done using a therometer
         AddSample (185, -18.0);  //    Temperature of inside my ice box
         AddSample (226,   0.0);  //    Melting ice in a cup of water
         AddSample (383,  18.2);  //    Ambient temperature in my shed
         AddSample (579,  36.0);  //    Body temperature
      }                           // By far not perfect, but good enough!

      ~t_Thermometer ()
      {
      }
      
      int GetRawValue (bool &Error)
      {
         int RawValue = analogRead (oPin);
         Error = (RawValue <= 5) || (RawValue > 1020); // Hardware error for values outside this range 
         return RawValue;
      }

      float RawValueToTemperature (unsigned long RawValue)
      {
         t_pSample pSample;
         float      Temp;
         int i;

         for (i=1; i<oSamples; i++) // Do not start at sample 0; sample 1 is the one that holds the slope and offset of the first segment
         {
            pSample = &oSampleArr[i];
            if (RawValue <= pSample->RawValue)
               break;
         }
         Temp = RawValue * pSample->Slope + pSample->Offset;
         return Temp;
      }

      // GetAverageTemperature returns the temperature in degree Celsius.
      // The function doesn't return immediately but takes 
      //    (SamplesPerMeasure-1)*DelayBetweenSamples
      // milliseconds to complete the measurement. If Error is true upon
      // return it means that there's probably a hardware error.

      float GetAverageTemperature (bool &Error)
      {
         unsigned int  RawValue;
         unsigned long Sum=0;

         for (int i=0; i<SamplesPerMeasure; i++)
         {
            if (i)
               delay (DelayBetweenSamples);
            RawValue = GetRawValue(Error);
            if (Error)
               return -DBL_MAX;
            Sum += RawValue;
         }
         return RawValueToTemperature (Sum/SamplesPerMeasure);
      }

   private:
      int      oPin;
      t_Sample oSampleArr[20]; // Max number of samples
      int      oSamples;
};

